"""
Selectors improve ``.loc[]`` indexing for multi-index pandas data.
"""

from functools import reduce
from operator import and_
from typing import Any, Mapping, Optional, Union

import numpy as np
from attrs import define, field
from pandas import DataFrame, Index, Series

from .types import Data
from .utils import shell_pattern_to_regex


def maybe_const(x):
    return x if isinstance(x, Selector) else Const(x)


class Selector:
    # Tell numpy that we want precedence
    __array_ufunc__ = None

    def __invert__(self):
        return Not(self)

    def __and__(self, other):
        if isinstance(other, Special):
            return other & self
        return And(self, maybe_const(other))

    def __or__(self, other):
        if isinstance(other, Special):
            return other | self
        return Or(self, maybe_const(other))

    __rand__ = __and__
    __ror__ = __or__


@define
class BinOp(Selector):
    a: Selector
    b: Selector


@define
class Const(Selector):
    val: Any

    def __call__(self, _):
        return self.val


@define
class And(BinOp):
    def __call__(self, df):
        return self.a(df) & self.b(df)


@define
class Or(BinOp):
    def __call__(self, df):
        return self.a(df) | self.b(df)


@define
class Not(Selector):
    a: Selector

    def __call__(self, df):
        return ~self.a.__call__(df)


class Special(Selector):
    pass


@define
class AllSelector(Special):
    def __invert__(self):
        return NoneSelector()

    def __and__(self, other):
        return other

    def __or__(self, other):
        return self

    def __call__(self, df):
        return Series(True, df.index)


@define
class NoneSelector(Special):
    def __invert__(self):
        return AllSelector()

    def __and__(self, other):
        return self

    def __or__(self, other):
        return other

    __rand__ = __and__
    __ror__ = __or__

    def __call__(self, df):
        return Series(False, df.index)


# Singletons for easy access
All = AllSelector()
None_ = NoneSelector()


@define
class Isin(Selector):
    filters: Mapping[str, Any]
    ignore_missing_levels: bool = field(default=False, repr=False)

    def __call__(self, df):
        if isinstance(df, Index):
            index = df
        else:
            index = df.index

        filters = self.filters
        if self.ignore_missing_levels:
            filters = {k: v for k, v in filters.items() if k in index.names}

        def apply_filter(value, level):
            if callable(value):
                return value(index.get_level_values(level))
            return index.isin(np.atleast_1d(value), level=level)

        return Series(
            reduce(and_, (apply_filter(v, k) for k, v in filters.items())), index
        )


def isin(
    df: Optional[Data] = None, ignore_missing_levels: bool = False, **filters: Any
) -> Union[Isin, Series]:
    """Constructs a MultiIndex selector.

    Arguments
    ---------
    df : Data, optional
        Data on which to match, if missing an ``Isin`` object is returned
    ignore_missing_levels : bool, default False
        If set, levels missing in data index will be ignored
    **filters
        Filter to apply on given levels (lists are ``or`` ed, levels are ``and`` ed)
        Callables are evaluated on the index level values.

    Returns
    -------
    Isin or Series

    Example
    -------
    >>> df.loc[isin(region="World", gas=["CO2", "N2O"])]

    or with explicit df to get a boolean mask

    >>> isin(df, region="World", gas=["CO2", "N2O"])
    """

    tester = Isin(filters, ignore_missing_levels=ignore_missing_levels)
    return tester if df is None else tester(df)


@define
class Ismatch(Selector):
    filters: Mapping[str, Any]
    regex: bool = False
    ignore_missing_levels: bool = field(default=False, repr=False)

    def index_match(self, index, patterns):
        matches = np.zeros((len(index),), dtype=bool)
        for pat in patterns:
            if isinstance(pat, str):
                if not self.regex:
                    pat = shell_pattern_to_regex(pat) + "$"
                matches |= index.str.match(pat, na=False)
            else:
                matches |= index == pat
        return matches

    def multiindex_match(self, index, patterns, level):
        level_num = index.names.index(level)
        levels = index.levels[level_num]

        matches = self.index_match(levels, patterns)

        (indices,) = np.where(matches)
        return np.in1d(index.codes[level_num], indices)

    def __call__(self, df):
        if isinstance(df, Index):
            index = df
        else:
            index = df.index

        single = self.filters.get(None) if len(self.filters) == 1 else None
        if single is not None:
            matches = self.index_match(index, np.atleast_1d(single))
        else:
            filters = self.filters
            if self.ignore_missing_levels:
                filters = {k: v for k, v in filters.items() if k in index.names}

            tests = (
                self.multiindex_match(index, np.atleast_1d(v), level=k)
                for k, v in filters.items()
            )
            matches = reduce(and_, tests)

        return Series(matches, index=index)


def ismatch(
    df: Union[None, Index, DataFrame, Series, str] = None,
    singlefilter: Optional[str] = None,
    regex: bool = False,
    ignore_missing_levels: bool = False,
    **filters,
) -> Union[Ismatch, Series]:
    """Constructs an Index or MultiIndex selector based on pattern matching.

    Arguments
    ---------
    df : Data, optional
        Data on which to match, if missing an ``Isin`` object is returned.
    singlefilter : str, optional
        Filter to apply on a non-multiindex index (can also be handed into the ``df``
        argument)
    regex : bool, default False
        If set, filters are interpreted as plain regex strings, otherwise (by default) a
        glob-like syntax is used
    ignore_missing_levels : bool, default False
        If set, levels missing in data index will be ignored
    **filters
        Filter to apply on given levels (lists are ``or`` ed, levels are ``and`` ed)

    Returns
    -------
    Isin or Series

    Example
    -------
    for a multiindex:

    >>> df.loc[ismatch(variable="Emissions|*|Fossil Fuel and Industry")]

    for a single index:

    >>> df.loc[ismatch("*bla*")]
    """

    if not filters and singlefilter is not None:
        filters = {None: singlefilter}

    if df is None:
        return Ismatch(
            filters, regex=regex, ignore_missing_levels=ignore_missing_levels
        )
    elif not isinstance(df, (DataFrame, Series, Index)):
        # Special case: a pattern was passed in through `df` which wants to be applied to
        # hopefully a non-MultiIndex based Series or dataframe that we get afterwards
        filters = {None: df}
        return Ismatch(filters, regex=regex)
    else:
        return Ismatch(
            filters, regex=regex, ignore_missing_levels=ignore_missing_levels
        )(df)



"""
Configuration file for then Sphinx documentation builder.
"""


# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
from importlib.metadata import version


# -- Project information --------------------------------------------------------------

source_suffix = ".rst"
master_doc = "index"
project = "pandas-indexing"
year = "2020"
author = ", ".join(["Jonas Hörsch"])
copyright = f"{year}, {author}"

# Retrieve package version from installed metadata
release = version("pandas-indexing")
version = ".".join(release.split(".")[:3])


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.autosectionlabel",
    # "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    # "sphinx.ext.todo",
    # "sphinx.ext.viewcode",
    "myst_nb",
]

if os.getenv("SPELLCHECK"):
    extensions += ("sphinxcontrib.spelling",)
    spelling_show_suggestions = True
    spelling_lang = "en_US"
    # https://sphinxcontrib-spelling.readthedocs.io/en/latest/customize.html
    spelling_word_list_filename = ["spelling_wordlist.txt"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["."]


extlinks = {
    "issue": (
        "https://github.com/coroa/pandas-indexing/issues/%s",
        "GH%s",
    ),
    "pull": ("https://github.com/coroa/pandas-indexing/pull/%s", "PR%s"),
}

# codecov io closes connection if host is accessed too repetitively.
# codecov links are ignored here for the same reason there's a sleep
# in the .travis.yml file
# see https://github.com/codecov/codecov-python/issues/158
linkcheck_ignore = [
    "https://codecov.io/gh/coroa/pandas-indexing/*",
]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["html", "jupyter_execute"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output -------------------------------------------------

html_theme = "sphinx_rtd_theme"
html_theme_path = ["_static"]

html_context = {
    "display_github": False,
    "github_user": "coroa",
    "github_repo": "pandas-indexing",
    "github_version": "main",
    "conf_py_path": "/docs/source",
}

html_static_path = ["_static"]
html_css_files = ["custom.css"]


# -- Extension configuration -------------------------------------------------

# -- Options for coverage extension ------------------------------------------
coverage_write_headline = False  # do not write headlines.

# -- Options for autodoc extension -------------------------------------------

# Do not add module names in the doc to hide the internal package structure of SeisBench
add_module_names = False
autodoc_default_options = {
    "members": True,
    "undoc-members": False,
    "private-members": False,
    "special-members": False,
    "inherited-members": True,
    "show-inheritance": True,
}

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "python": ("https://docs.python.org/3", None),
    "pyam": ("https://pyam-iamc.readthedocs.io/en/latest", None),
    "scmdata": ("https://scmdata.readthedocs.io/en/latest", None),
    # "pint": ("https://pint.readthedocs.io/en/latest", None), # no full API doc here, unfortunately
}

# -- Options for napoleon extension ------------------------------------------
napoleon_google_docstring = False
napoleon_numpy_docstring = True
set_type_checking_flag = False

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True



"""
Test datasets for simplifying the tutorial.
"""

from importlib.resources import open_text

import pandas as pd


def remindhighre_power():
    """Read IAMC power sector data from REMIND's HighRE IMP scenario for AR6.

    Returns
    -------
    pd.DataFrame
        Time-series like power sector data

    .. [1] Luderer, G., Madeddu, S., Merfort, L. et al. Impact of declining renewable
       energy costs on electrification in low-emission scenarios. Nat Energy 6, 32–42
       (2022). https://doi.org/10.1038/s41560-021-00937-z
    """
    return pd.read_csv(
        open_text(__name__, "remindhighre_power.csv"), index_col=list(range(5))
    ).rename(columns=int)



"""Provide aligned basic arithmetic ops.

Simple arithmetic operations :py:func:`add`, :py:func:`divide`, :py:func:`multiply` and
:py:func:`subtract` which allow setting the standard how="outer" alignment that pandas
uses by default.

In practice, this means if dataframes do not share the same axes one can choose to get
the results for only the items index items existing in both indices (``how="inner"``) or
whether to prefer the axis from the first (``how="left"``) or the right (``how="right``)
operand.

See also
--------
pandas.DataFrame.align
"""

import operator
from typing import Any, Dict, Optional

from pandas import DataFrame, Series
from pandas.core.ops import ARITHMETIC_BINOPS

from .core import assignlevel, uniquelevel
from .types import Axis, Data


ALTERNATIVE_NAMES = {
    "truediv": ["div", "divide"],
    "rtruediv": ["rdiv"],
    "mul": ["multiply"],
    "sub": ["subtract"],
}

ARITHMETIC_UNITBINOPS = [
    "add",
    "radd",
    "sub",
    "rsub",
    "mod",
    "mul",
    "rmul",
    "truediv",
    "rtruediv",
]


def _needs_axis(df: Data, other: Data) -> bool:
    return (isinstance(df, DataFrame) and isinstance(other, Series)) or (
        isinstance(df, Series) and isinstance(other, DataFrame)
    )


def _create_binop(op: str):
    def binop(
        df: Data,
        other: Data,
        assign: Optional[Dict[str, Any]] = None,
        axis: Optional[Axis] = None,
        **align_kwargs: Any,
    ):
        if assign is not None:
            df = assignlevel(df, **assign)
            other = assignlevel(other, **assign)

        align_kwargs.setdefault("copy", False)
        if _needs_axis(df, other):
            axis = 0
        if isinstance(df, Series) and isinstance(other, DataFrame):
            if align_kwargs.get("join") in ("left", "right"):
                align_kwargs["join"] = {"left": "right", "right": "left"}[
                    align_kwargs["join"]
                ]
            other, df = other.align(df, axis=axis, **align_kwargs)
        else:
            df, other = df.align(other, axis=axis, **align_kwargs)

        return getattr(df, op)(other, axis=axis)

    return binop


def _create_unitbinop(op, binop):
    def unitbinop(
        df: Data,
        other: Data,
        level: str = "unit",
        assign: Optional[Dict[str, Any]] = None,
        axis: Optional[Axis] = None,
        **align_kwargs: Any,
    ):
        unitaxis = 0 if axis is None else axis
        df_unit = uniquelevel(df, level, axis=unitaxis).item()
        other_unit = uniquelevel(other, level, axis=unitaxis).item()

        import pint

        ur = pint.get_application_registry()
        if op in ("add", "radd", "sub", "rsub", "mod"):
            quantity = ur(df_unit)
            other = other * ur(other_unit).to(quantity.units).m
        elif op in ("mul", "truediv"):
            quantity = getattr(operator, op)(
                ur(df_unit), ur(other_unit)
            ).to_reduced_units()
        elif op in ("rmul", "rtruediv"):
            quantity = getattr(operator, op[1:])(
                ur(other_unit), ur(df_unit)
            ).to_reduced_units()
        else:
            raise NotImplementedError()

        if assign is None:
            assign = dict()
        assign = {level: f"{quantity.units:~}"} | assign

        return binop(df, other, assign=assign, axis=axis, **align_kwargs) * quantity.m

    return unitbinop


for op in ARITHMETIC_BINOPS:
    names = [op] + ALTERNATIVE_NAMES.get(op, [])
    binop = _create_binop(op)
    globals().update({name: binop for name in names})
    if op in ARITHMETIC_UNITBINOPS:
        unitbinop = _create_unitbinop(op, binop)
        globals().update({f"unit{name}": unitbinop for name in names})



"""Registers convenience accessors into the ``pix`` namespace of each pandas
object.

Examples
--------
>>> df.pix.project(["model", "scenario"])

>>> df.index.pix.assign(unit="Mt CO2")

>>> df.pix.multiply(other, how="left")
"""

import warnings
from typing import Any, Callable, Dict, Literal, Mapping, Optional, Sequence, Union

import pandas as pd
from deprecated.sphinx import deprecated
from pandas import DataFrame, Index, MultiIndex, Series
from pandas.api.extensions import no_default

from . import arithmetics
from .core import (
    add_zeros_like,
    aggregatelevel,
    antijoin,
    assignlevel,
    describelevel,
    dropnalevel,
    extractlevel,
    fixindexna,
    formatlevel,
    isna,
    notna,
    projectlevel,
    semijoin,
    to_tidy,
    uniquelevel,
)
from .types import Axis, Data
from .units import convert_unit, dequantify, quantify
from .utils import doc


class _PixAccessor:
    """
    Convenience accessor for accessing `pandas-indexing` functions.
    """

    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def __repr__(self):
        return describelevel(self._obj, as_str=True)

    @doc(assignlevel, df="")
    def assign(
        self,
        frame: Optional[Data] = None,
        order: bool = False,
        axis: Axis = 0,
        ignore_index: bool = False,
        **labels: Any,
    ) -> Union[DataFrame, Series, MultiIndex]:
        return assignlevel(
            self._obj,
            frame=frame,
            order=order,
            axis=axis,
            ignore_index=ignore_index,
            **labels,
        )

    @doc(extractlevel, index_or_data="")
    def extract(
        self,
        template: Optional[str] = None,
        *,
        keep: bool = False,
        dropna: bool = True,
        regex: bool = False,
        axis: Axis = 0,
        drop: Optional[bool] = None,
        **templates: str,
    ) -> Union[DataFrame, Series, Index]:
        if drop is not None:
            warnings.warn(
                "Argument `drop` is deprecated (use `keep` instead)", DeprecationWarning
            )
            keep = not drop

        return extractlevel(
            self._obj,
            template,
            keep=keep,
            dropna=dropna,
            regex=regex,
            axis=axis,
            **templates,
        )

    @doc(formatlevel, index_or_data="")
    def format(
        self, axis: Axis = 0, **templates: str
    ) -> Union[DataFrame, Series, Index]:
        return formatlevel(self._obj, axis=axis, **templates)

    @doc(uniquelevel, index_or_data="")
    def unique(
        self,
        levels: Union[str, Sequence[str], None],
        axis: Axis = 0,
    ) -> Index:
        return uniquelevel(self._obj, levels=levels, axis=axis)

    @doc(projectlevel, index_or_data="")
    def project(
        self,
        levels: Sequence[str],
        axis: Axis = 0,
    ) -> Union[DataFrame, Series, Index]:
        return projectlevel(self._obj, levels=levels, axis=axis)

    @doc(notna, index_or_data="")
    def notna(
        self,
        subset: Optional[Sequence[str]] = None,
        how: Literal["any", "all"] = "any",
        axis: Axis = 0,
    ):
        return notna(self._obj, subset=subset, how=how, axis=axis)

    @doc(isna, index_or_data="")
    def isna(
        self,
        subset: Optional[Sequence[str]] = None,
        how: Literal["any", "all"] = "any",
        axis: Axis = 0,
    ):
        return isna(self._obj, subset=subset, how=how, axis=axis)

    @doc(dropnalevel, index_or_data="")
    def dropna(
        self,
        subset: Optional[Sequence[str]] = None,
        how: Literal["any", "all"] = "any",
        axis: Axis = 0,
    ) -> Union[DataFrame, Series, Index]:
        return dropnalevel(self._obj, subset=subset, how=how, axis=axis)

    @doc(fixindexna, index_or_data="")
    def fixna(
        self,
        axis: Axis = 0,
    ) -> Union[DataFrame, Series, Index]:
        return fixindexna(self._obj, axis=axis)

    @doc(antijoin, index_or_data="")
    def antijoin(self, other: Index, *, axis: Axis = 0):
        return antijoin(self._obj, other, axis=axis)


class _DataPixAccessor(_PixAccessor):
    @doc(semijoin, frame_or_series="")
    def semijoin(
        self,
        other: Index,
        *,
        how: Literal["left", "right", "inner", "outer"] = "left",
        level: Union[str, int, None] = None,
        sort: bool = False,
        axis: Axis = 0,
        fill_value: Any = no_default,
        fail_on_reorder: bool = False,
    ) -> Union[DataFrame, Series]:
        return semijoin(
            self._obj,
            other,
            how=how,
            level=level,
            sort=sort,
            axis=axis,
            fill_value=fill_value,
            fail_on_reorder=fail_on_reorder,
        )

    @doc(quantify, data="", example_call="s.pix.quantify()")
    def quantify(
        self,
        level: str = "unit",
        unit: Optional[str] = None,
        axis: Axis = 0,
        copy: bool = False,
    ):
        return quantify(self._obj, level=level, unit=unit, axis=axis, copy=copy)

    def dequantify(self, level: str = "unit", axis: Axis = 0, copy: bool = False):
        return dequantify(self._obj, level=level, axis=axis, copy=copy)

    @doc(
        convert_unit,
        data="",
        convert_unit_s_km='s.pix.convert_unit("km")',
        convert_unit_s_m_to_km='s.pix.convert_unit({"m": "km"})',
    )
    def convert_unit(
        self,
        unit: Union[str, Mapping[str, str], Callable[[str], str]],
        level: Optional[str] = "unit",
        axis: Axis = 0,
    ):
        return convert_unit(self._obj, unit=unit, level=level, axis=axis)

    @doc(to_tidy, data="")
    def to_tidy(
        self,
        meta: Optional[DataFrame] = None,
        value_name: Optional[str] = "value",
        columns: Optional[str] = "year",
    ):
        return to_tidy(self._obj, meta=meta, value_name=value_name, columns=columns)

    @doc(aggregatelevel, data="")
    def aggregate(
        self,
        agg_func: str = "sum",
        axis: Axis = 0,
        dropna: bool = True,
        mode: Literal["replace", "append", "return"] = "replace",
        **levels: Dict[str, Sequence[Any]],
    ):
        return aggregatelevel(
            self._obj, agg_func=agg_func, axis=axis, dropna=dropna, mode=mode, **levels
        )

    @doc(add_zeros_like, data="")
    def add_zeros_like(
        self,
        reference: Union[MultiIndex, DataFrame, Series],
        /,
        derive: Optional[Dict[str, MultiIndex]] = None,
        **levels: Sequence[str],
    ):
        return add_zeros_like(self._obj, reference=reference, derive=derive, **levels)


def _create_forward_binop(op):
    def forward_binop(
        self,
        other: Data,
        assign: Optional[Dict[str, Any]] = None,
        axis: Optional[Axis] = None,
        **align_kwargs: Any,
    ):
        return getattr(arithmetics, op)(
            self._obj, other, assign=assign, axis=axis, **align_kwargs
        )

    return forward_binop


def _create_forward_unitbinop(op):
    def forward_unitbinop(
        self,
        other: Data,
        level: str = "unit",
        assign: Optional[Dict[str, Any]] = None,
        axis: Optional[Axis] = None,
        **align_kwargs: Any,
    ):
        return getattr(arithmetics, f"unit{op}")(
            self._obj, other, level=level, assign=assign, axis=axis, **align_kwargs
        )

    return forward_unitbinop


for op in arithmetics.ARITHMETIC_BINOPS:
    forward_binop = _create_forward_binop(op)
    setattr(_DataPixAccessor, op, forward_binop)
    for alt in arithmetics.ALTERNATIVE_NAMES.get(op, []):
        setattr(_DataPixAccessor, alt, forward_binop)

for op in arithmetics.ARITHMETIC_UNITBINOPS:
    forward_unitbinop = _create_forward_unitbinop(op)
    setattr(_DataPixAccessor, f"unit{op}", forward_unitbinop)
    for alt in arithmetics.ALTERNATIVE_NAMES.get(op, []):
        setattr(_DataPixAccessor, f"unit{alt}", forward_unitbinop)


@pd.api.extensions.register_dataframe_accessor("pix")
class DataFramePixAccessor(_DataPixAccessor):
    pass


@pd.api.extensions.register_series_accessor("pix")
class SeriesPixAccessor(_DataPixAccessor):
    pass


@pd.api.extensions.register_index_accessor("pix")
class IndexPixAccessor(_PixAccessor):
    pass


use_pix_instead = deprecated(
    reason="Use the new name ``df.pix`` of the accessor", version="0.2.9"
)


@pd.api.extensions.register_dataframe_accessor("idx")
@use_pix_instead
class DataFrameIdxAccessor(_DataPixAccessor):
    pass


@pd.api.extensions.register_series_accessor("idx")
@use_pix_instead
class SeriesIdxAccessor(_DataPixAccessor):
    pass


@pd.api.extensions.register_index_accessor("idx")
@use_pix_instead
class IndexIdxAccessor(_PixAccessor):
    pass



"""Unit handling in pandas data.

Enables unit conversions based on pint's application registry (see also Notes).

By default units are expected -- as in the IAMC default format -- on a ``unit`` level on
each row, but a column-wise ``unit`` level is also supported.

Units can be handled in one of two flavours:

1. :py:func:`convert_unit` converts manually to a new unit like `convert_unit(s, "km")`

2. :py:func:`quantify` convert data to a pint pandas array which tracks units implicitly
   through arithmetics until :py:func:`dequantify` then extracts the tracked unit back
   into the multiindex level.

   While this is in theory the simpler approach, the underlying library ``pint-pandas`` [1]_
   is brittle and breaks from time to time.

Notes
-----
The pint application registry is set by :py:func:`pint.set_application_registry` or
with :py:func:`set_openscm_registry_as_default`. The latter sets the IAMC based ``openscm-units`` one [2]_.

Examples
--------
>>> import pandas_indexing as pi
>>> pi.set_openscm_registry_as_default()
>>> s = Series(
...     [7, 8],
...     MultiIndex.from_tuples([("foo", "mm"), ("bar", "m")], names=["var", "unit"]),
... )
>>> s = pi.convert_unit(s, "km")
>>> s
var  unit
bar  km      0.008000
foo  km      0.000007
dtype: float64

>>> pi.quantify(s)
var
bar    0.008
foo    7e-06
dtype: pint[kilometer]

References
----------
.. [1] https://github.com/hgrecco/pint-pandas
.. [2] https://github.com/openscm/openscm-units

See also
--------
pint.set_application_registry
"""

from typing import Callable, Mapping, Optional, Union

from pandas import DataFrame, Series


INSTALL_PACKAGE_WARNING = "Please install {package} via conda or pip, or use the pandas-indexing[units] variant."

try:
    from pint_pandas import PintType

    has_pint_pandas = True
except ImportError:
    has_pint_pandas = False

try:
    import pint

    has_pint = True
except ImportError:
    has_pint = False


from .core import assignlevel, uniquelevel
from .types import Axis, Data
from .utils import doc


@doc(
    data="""
    data : DataFrame or Series
        DataFrame or Series to quantify\
    """,
    example_call="quantify(s)",
)
def quantify(
    data: Data,
    level: str = "unit",
    unit: Optional[str] = None,
    axis: Axis = 0,
    copy: bool = False,
) -> Data:
    """Convert columns in `data` to pint extension types to handle units.

    `pint-pandas <https://github.com/grecco/pint-pandas>`_ can only represent a single
    unit per column and is somewhat brittle.

    Parameters
    ----------\
    {data}
    unit : str, optional
        If given, assumes data is currently in this unit.
    level : str, optional
        Level of which to use the unit, by default "unit"
    axis : Axis, optional
        Axis from which to pop the `level`, by default 0
    copy : bool, optional
        Whether data should be copied, by default False

    Returns
    -------
    Data
        Data with internalized unit which stays with arithmetics

    Raises
    ------
    ValueError
        If `level` contains more than one unit

    Examples
    --------
    >>> s = Series(
    ...     [7e-3, 8],
    ...     MultiIndex.from_tuples([("foo", "m"), ("bar", "m")], names=["var", "unit"]),
    ... )
    >>> {example_call}
    var
    foo    7e-06
    bar    0.008
    dtype: pint[kilometer]

    Notes
    -----
    pint-pandas uses the pint application registry, which can be set with
    :py:func:`pint.set_application_registry` or
    :py:func:`set_openscm_registry_as_default`.

    See also
    --------
    set_openscm_registry_as_default
    dequantify
    convert_unit
    """
    assert has_pint_pandas, INSTALL_PACKAGE_WARNING.format(package="pint-pandas")

    if unit is None:
        unit = uniquelevel(data, level, axis)
        if len(unit) != 1:
            raise ValueError(
                f"pint-pandas can only represent homogeneous units: {', '.join(unit)}"
            )
        unit = unit[0]
        data = data.droplevel(level, axis=axis)

    return data.astype(PintType(unit), copy=copy)


def format_dtype(dtype):
    if not isinstance(dtype, PintType):
        return ""

    return ("{:" + dtype.ureg.default_format + "}").format(dtype.units)


def dequantify(data: Data, level: str = "unit", axis: Axis = 0, copy: bool = False):
    assert has_pint_pandas, INSTALL_PACKAGE_WARNING.format(package="pint-pandas")

    if isinstance(data, Series):
        unit = format_dtype(data.dtype)
        data = data.pint.magnitude
    elif isinstance(data, DataFrame):
        if axis in (0, "index"):
            unit = data.dtypes.unique()
            if len(unit) != 1:
                raise ValueError(
                    f"Only homogeneous units can be represented: {', '.join(unit)}"
                )
            unit = format_dtype(unit[0])
            data = data.apply(lambda s: s.pint.magnitude, result_type="expand", axis=1)
        elif axis in (1, "columns"):
            unit = data.dtypes.map(format_dtype)
            data = data.apply(lambda s: s.pint.magnitude, result_type="expand", axis=0)
        else:
            raise ValueError(
                f"axis can only be one of 0, 1, 'index' or 'columns', not: {axis}"
            )
    else:
        raise TypeError(
            f"data must derive from DataFrame or Series, but is {type(data)}"
        )

    return assignlevel(data, unit=unit, axis=axis)


@doc(
    data="""
    data : DataFrame or Series
        DataFrame or Series with a "unit" level\
    """,
    convert_unit_s_km='convert_unit(s, "km")',
    convert_unit_s_m_to_km='convert_unit(s, {"m": "km"})',
)
def convert_unit(
    data: Data,
    unit: Union[str, Mapping[str, str], Callable[[str], str]],
    level: Optional[str] = "unit",
    axis: Axis = 0,
):
    """Converts units in a dataframe or series.

    Parameters
    ----------\
    {data}
    unit : str or dict or function from old to new unit
        Either a single target unit or a mapping from old unit to target unit
        (a unit missing from the mapping or with a return value of None is kept)
    level : str|None, default "unit"
        Level name on ``axis``
        If None, then ``unit`` needs to be a mapping like ``{{from_unit: to_unit}}``
    axis : Axis, default 0
        Axis of unit level

    Returns
    -------
    Data
        DataFrame or Series with converted units

    Examples
    --------
    >>> s = Series(
    ...     [7, 8],
    ...     MultiIndex.from_tuples(
    ...         [("foo", "mm"), ("bar", "m")], names=["var", "unit"]
    ...     ),
    ... )
    >>> {convert_unit_s_km}
    var  unit
    bar  km      0.008000
    foo  km      0.000007
    dtype: float64

    >>> {convert_unit_s_m_to_km}
    var  unit
    bar  km      0.008
    foo  mm      7.000
    dtype: float64

    Notes
    -----
    Uses the pint application registry, which can be set with
    :py:func:`pint.set_application_registry` or
    :py:func:`set_openscm_registry_as_default`.

    See also
    --------
    set_openscm_registry_as_default
    quantify
    dequantify
    """
    if callable(unit):
        unit_map = unit
    elif isinstance(unit, dict):
        unit_map = unit.get
    else:
        unit_map = lambda u: unit

    ur = pint.get_application_registry()

    def _convert_unit(df, old_unit=None):
        if old_unit is None:
            old_unit = df.name
        new_unit = unit_map(old_unit)
        if new_unit is None:
            return df

        factor = ur.Quantity(1, old_unit).to(new_unit).m
        df = factor * df
        if level is None:
            return df
        return assignlevel(df, **{level: new_unit})

    try:
        if level is None:
            if not (isinstance(unit, dict) and len(unit) == 1):
                raise ValueError(
                    "If level is None, unit must look like {{fromunit: tounit}}"
                )
            (old_unit,) = unit.keys()
            data = _convert_unit(data, old_unit=old_unit)
        elif axis in (0, "index"):
            data = data.groupby(level=level, group_keys=False).apply(_convert_unit)
        elif axis in (1, "columns"):
            data = data.T.groupby(level=level, group_keys=False).apply(_convert_unit).T
        else:
            raise ValueError(
                f"axis can only be one of 0, 1, 'index' or 'columns', not: {axis}"
            )

        return data.__finalize__(data, "convert_unit")
    except pint.errors.DimensionalityError as exc:
        raise exc from None  # remove exception chaining


_openscm_registry = None


def get_openscm_registry(add_co2e: bool = True):
    global _openscm_registry
    if _openscm_registry is not None:
        return _openscm_registry

    import openscm_units

    if add_co2e:
        _openscm_registry = openscm_units.ScmUnitRegistry()
        _openscm_registry.add_standards()
        _openscm_registry.define("CO2e = CO2")
        _openscm_registry.define("CO2eq = CO2")
    else:
        _openscm_registry = openscm_units.unit_registry

    return _openscm_registry


def set_openscm_registry_as_default(add_co2e: bool = True):
    unit_registry = get_openscm_registry(add_co2e=add_co2e)

    assert has_pint, INSTALL_PACKAGE_WARNING.format(package="pint")
    app_registry = pint.get_application_registry()
    if unit_registry is not app_registry:
        pint.set_application_registry(unit_registry)

    return unit_registry


def is_unit(unit: str) -> bool:
    assert has_pint, INSTALL_PACKAGE_WARNING.format(package="pint")
    ur = pint.get_application_registry()
    try:
        ur.Unit(unit)
        return True
    except pint.errors.UndefinedUnitError:
        return False



from typing import Literal, TypeVar, Union

from pandas import DataFrame, Index, Series


Axis = Literal[0, 1, "index", "columns"]
Data = Union[Series, DataFrame]
T = TypeVar("T", bound=Union[Index, DataFrame, Series])
S = TypeVar("S", bound=Union[DataFrame, Series])



"""Utils module.

Simple utility functions not of greater interest
"""

import re
from typing import Union

from pandas import DataFrame, Index, Series
from pandas.util._decorators import doc  # noqa: F401

from .types import Axis


def shell_pattern_to_regex(s):
    """
    Escape characters with specific regexp use.
    """
    return re.escape(s).replace(r"\*\*", ".*").replace(r"\*", r"[^|]*")


def print_list(x, n):
    """Return a printable string of a list shortened to n characters.

    Copied from pyam.utils.print_list by Daniel Huppmann, licensed under
    Apache 2.0.

    https://github.com/IAMconsortium/pyam/blob/449b77cb1c625b455e3675801477f19e99b30e64/pyam/utils.py#L599-L638
    .
    """
    # if list is empty, only write count
    if len(x) == 0:
        return "(0)"

    # write number of elements, subtract count added at end from line width
    x = [i if i != "" else "''" for i in map(str, x)]
    count = f" ({len(x)})"
    n -= len(count)

    # if not enough space to write first item, write shortest sensible line
    if len(x[0]) > n - 5:
        return "..." + count

    # if only one item in list
    if len(x) == 1:
        return f"{x[0]} (1)"

    # add first item
    lst = f"{x[0]}, "
    n -= len(lst)

    # if possible, add last item before number of elements
    if len(x[-1]) + 4 > n:
        return lst + "..." + count
    else:
        count = f"{x[-1]}{count}"
        n -= len({x[-1]}) + 3

    # iterate over remaining entries until line is full
    for i in x[1:-1]:
        if len(i) + 6 <= n:
            lst += f"{i}, "
            n -= len(i) + 2
        else:
            lst += "... "
            break

    return lst + count


def get_axis(data: Union[Index, Series, DataFrame], axis: Axis = 0):
    """
    Get axis `axis` from `data`
    """
    if isinstance(data, Index):
        return data
    elif isinstance(data, Series):
        return data.index
    elif isinstance(data, DataFrame):
        if axis in (0, "index"):
            return data.index
        elif axis in (1, "columns"):
            return data.columns
        else:
            raise ValueError(
                f"axis can only be one of 0, 1, 'index' or 'columns', not: {axis}"
            )
    else:
        raise ValueError(
            f"data needs to be a pandas Series or DataFrame, not: {type(data)}"
        )


def quote_list(l):
    return ", ".join(f'"{s}"' for s in l)


def s(l):
    return "s" if len(l) > 1 else ""



"""
Pandas indexing helper.
"""

from importlib.metadata import version as _version

from . import accessors, core, datasets
from .arithmetics import add, divide, multiply, subtract
from .core import (
    aggregatelevel,
    antijoin,
    assignlevel,
    concat,
    describelevel,
    dropnalevel,
    extractlevel,
    formatlevel,
    index_names,
    projectlevel,
    semijoin,
    to_tidy,
    uniquelevel,
)
from .selectors import All, None_, isin, ismatch
from .units import convert_unit, dequantify, quantify, set_openscm_registry_as_default


try:
    __version__ = _version("pandas-indexing")
except Exception:
    # Local copy or not installed with setuptools.
    # Disable minimum version checks on downstream libraries.
    __version__ = "999"



"""
Core module.
"""

import re
import warnings
from functools import reduce
from itertools import chain, product
from operator import and_, or_
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
from deprecated import deprecated
from pandas import DataFrame, Index, MultiIndex, Series
from pandas.api.extensions import no_default
from pandas.core.indexes.frozen import FrozenList

from .selectors import isin
from .types import Axis, Data, S, T
from .utils import doc, get_axis, print_list, quote_list, s


def _assignlevel(
    index: Index,
    frame: Optional[Data] = None,
    order: bool = False,
    ignore_index: bool = False,
    **labels: Any,
) -> MultiIndex:
    if isinstance(index, MultiIndex):
        new_levels = list(index.levels)
        new_codes = list(index.codes)
        new_names = list(index.names)
    else:
        level = index.unique().dropna()
        new_levels = [level]
        new_codes = [level.get_indexer(index)]
        new_names = [index.name]

    if ignore_index:

        def maybe_align_with_index(df_or_ser):
            return df_or_ser

    else:

        def maybe_align_with_index(df_or_ser):
            if isinstance(df_or_ser, (DataFrame, Series)):
                return semijoin(df_or_ser, index, how="right", fail_on_reorder=True)

            return df_or_ser

    labels = {level: maybe_align_with_index(lbls) for level, lbls in labels.items()}

    if frame is not None:
        if isinstance(frame, Series):
            frame = frame.to_frame()
        if isinstance(frame, DataFrame):
            labels = dict(chain(maybe_align_with_index(frame).items(), labels.items()))
        else:
            raise ValueError(
                f"frame must be a DataFrame or Series, but is: {type(frame)}"
            )

    for level, lbls in labels.items():
        if np.isscalar(lbls):
            lvl = [lbls]
            cde = np.full(len(index), 0)
        else:
            lvl = Index(lbls).unique().dropna()
            cde = lvl.get_indexer(lbls)

        if level in new_names:
            i = new_names.index(level)
            new_levels[i] = lvl
            new_codes[i] = cde
        else:
            new_levels.append(lvl)
            new_codes.append(cde)
            new_names.append(level)

    new_index = MultiIndex(codes=new_codes, levels=new_levels, names=new_names)

    if order:
        new_index = new_index.reorder_levels(order)

    return new_index


@doc(
    df="""
    df : DataFrame, Series or Index
        Index, Series or DataFrame of which to change index levels\
    """
)
def assignlevel(
    df: T,
    frame: Optional[Data] = None,
    order: bool = False,
    axis: Axis = 0,
    ignore_index: bool = False,
    **labels: Any,
) -> T:
    """Add or overwrite levels on a multiindex.

    Parameters
    ----------\
    {df}
    frame : Series or DataFrame, optional
        Additional labels
    order : list of str, optional
        Level names in desired order or False, by default False
    axis : {{0, 1, "index", "columns"}}, default 0
        Axis where to update multiindex
    ignore_index : bool, optional
        If true, dataframes or series are not index aligned
    **labels
        Labels for each new index level

    Returns
    -------
    df
        Series or DataFrame with changed index or new MultiIndex
    """
    if isinstance(df, Index):
        return _assignlevel(
            df, frame=frame, order=order, ignore_index=ignore_index, **labels
        )

    index = get_axis(df, axis)
    new_index = _assignlevel(
        index, frame=frame, order=order, ignore_index=ignore_index, **labels
    )
    return df.set_axis(new_index, axis=axis)


def _projectlevel(index: Index, levels: Sequence[str]) -> Index:
    levels = np.atleast_1d(levels)

    missing_levels = [level for level in levels if level not in index.names]
    if missing_levels:
        raise KeyError(
            f"Index has no level{s(missing_levels)} {quote_list(missing_levels)} "
            f"(existing levels: {quote_list(index.names)})"
        )

    if len(levels) == 1:
        return index.get_level_values(levels[0])

    return index.droplevel(index.names.difference(levels)).reorder_levels(levels)


@doc(
    index_or_data="""
    index_or_data : DataFrame, Series or Index
        Index, Series or DataFrame to project\
    """
)
def projectlevel(index_or_data: T, levels: Sequence[str], axis: Axis = 0) -> T:
    """Project multiindex to given *levels*.

    Drops all levels except the ones explicitly mentioned from a given multiindex
    or an axis of a series or a dataframe.

    Parameters
    ----------\
    {index_or_data}
    levels : sequence of str
        Names of levels to project on (to keep)
    axis : {{0, 1, "index", "columns"}}, default 0
        Axis of DataFrame to project

    Returns
    -------
    index_or_data : Index|MultiIndex|Series|DataFrame

    See also
    --------
    pandas.MultiIndex.droplevel
    pandas.Series.droplevel
    pandas.DataFrame.droplevel
    """
    if isinstance(index_or_data, Index):
        return _projectlevel(index_or_data, levels)

    index = get_axis(index_or_data, axis)
    return index_or_data.set_axis(_projectlevel(index, levels), axis=axis)


def concat(
    objs: Union[Iterable[T], Mapping[str, T]],
    order: Optional[Sequence[str]] = None,
    axis: Axis = 0,
    keys: Union[None, str, Index, Sequence] = None,
    copy: bool = False,
    **concat_kwds,
) -> T:
    """Concatenate pandas objects along a particular axis.

    In addition to the functionality provided by pd.concat, if the concat axis has a multiindex
    then the level order is reordered consistently.

    Parameters
    ----------
    objs : a sequence or mapping of Series, DataFrame or Index objects
        If a mapping is passed the keys will be used as a new index level
        (with the name of the `keys` argument).
    order : a sequence of str, default None
        The order of level names in which to concatenate
    axis : Axis
        Axis along which to concatenate
    keys : str or list-like of str
        If `objs` is a mapping, a string-like value will be used as name of the new level,
        otherwise it is passed on to pd.concat.
    copy : bool, default False
        Whether to copy the underlying data
    **concat_kwds
        Other arguments accepted by pd.concat

    Returns
    -------
    Concatenated data or index

    Raises
    ------
    ValueError
        If the level names of `objs` do not match

    See also
    --------
    pandas.concat
    """

    if isinstance(objs, dict):
        objs = {k: v for k, v in objs.items() if v is not None}
        keys = Index(objs.keys(), name=keys)
        objs = list(objs.values())
    else:
        objs = [obj for obj in objs if obj is not None]

    if not objs:
        raise ValueError("Need at least one element to concatenate")

    if order is None:
        order = get_axis(objs[0], axis=axis).names
    elif isinstance(keys, Index):
        # make sure the order list does not include the new axis name
        order = [o for o in order if o != keys.name]

    orderset = frozenset(order)

    def reorder(df_ser_or_idx):
        ax = get_axis(df_ser_or_idx, axis=axis)
        if ax.names == order:
            return df_ser_or_idx
        elif not set(ax.names) == orderset:
            raise ValueError(
                "All objects need to have the same index levels, but "
                f"{set(orderset)} != {set(ax.names)}"
            )
        idx = ax.reorder_levels(order)
        if isinstance(df_ser_or_idx, Index):
            return idx

        return df_ser_or_idx.set_axis(idx, axis=axis, copy=False)

    objs = [reorder(o) for o in objs]
    if not isinstance(objs[0], Index):
        return pd.concat(objs, keys=keys, copy=copy, axis=axis, **concat_kwds)

    if keys is not None:
        keys = Index(keys)
        objs = [assignlevel(o, **{keys.name: k}) for o, k in zip(objs, keys)]

    return reduce(lambda x, y: x.append(y), objs)


def _notna(
    index: Index,
    subset: Optional[Sequence[str]] = None,
    how: Literal["any", "all"] = "any",
) -> np.ndarray:
    if not isinstance(index, MultiIndex):
        return index.notna()

    subset = index.names if subset is None else np.atleast_1d(subset)
    codes = [index.codes[index.names.index(n)] for n in subset]
    op = and_ if how == "any" else or_
    return reduce(op, [c != -1 for c in codes])


def notna(
    index_or_data: Union[Index, Series, DataFrame],
    subset: Optional[Sequence[str]] = None,
    how: Literal["any", "all"] = "any",
    axis: Axis = 0,
):
    return _notna(get_axis(index_or_data, axis), subset, how)


def isna(
    index_or_data: Union[Index, Series, DataFrame],
    subset: Optional[Sequence[str]] = None,
    how: Literal["any", "all"] = "any",
    axis: Axis = 0,
):
    return ~_notna(get_axis(index_or_data, axis), subset, how)


@doc(
    index_or_data="""
    index_or_data : DataFrame, Series or Index
        Index, Series or DataFrame of which to drop rows or columns\
    """
)
def dropnalevel(
    index_or_data: T,
    subset: Optional[Sequence[str]] = None,
    how: Literal["any", "all"] = "any",
    axis: Axis = 0,
) -> T:
    """Remove missing index values.

    Drops all index entries for which any or all (``how``) levels are
    undefined.

    Parameters
    ----------\
    {index_or_data}
    subset : Sequence[str], optional
        Names of levels on which to check for NA values
    how : {{"any", "all"}}
        Whether to remove an entry if all levels are NA or only a single one
    axis : {{0, 1, "index", "columns"}}, default 0
        Axis of DataFrame to check on

    Returns
    -------
    index_or_data : Index|MultiIndex|Series|DataFrame

    See also
    --------
    pandas.DataFrame.dropna
    pandas.Series.dropna
    pandas.Index.dropna
    """
    if isinstance(index_or_data, Index):
        return index_or_data[_notna(index_or_data, subset, how)]

    if axis in (0, "index"):
        return index_or_data.loc[_notna(index_or_data.index, subset, how)]

    return index_or_data.loc[:, _notna(index_or_data.columns, subset, how)]


@doc(
    index_or_data="""
    index_or_data : DataFrame, Series or Index
        Index, Series or DataFrame of which to describe index levels\
    """
)
def uniquelevel(
    index_or_data: Union[DataFrame, Series, Index],
    levels: Union[str, Sequence[str], None],
    axis: Axis = 0,
) -> Index:
    """Return unique index levels.

    Parameters
    ----------\
    {index_or_data}
    levels : str or Sequence[str], optional
        Names of levels to get unique values of
    axis : {{0, 1, "index", "columns"}}, default 0
        Axis of DataFrame to check on

    Returns
    -------
    unique_index : Index

    See also
    --------
    pandas.Index.unique
    """
    index = get_axis(index_or_data, axis)

    if levels is None or isinstance(levels, str):
        return index.unique(level=levels)

    levels = list(levels)
    if len(levels) == 1:
        return index.unique(level=levels[0])

    return projectlevel(index, levels).unique()


def _describelevel(index: Index, n: int = 80) -> str:
    def name(l):
        return "<unnamed>" if l is None else l

    c1 = max(len(name(l)) for l in index.names) + 1
    c2 = n - c1 - 5
    return "\n".join(
        f" * {name(l):{c1}}: {print_list(index.unique(l), c2)}" for l in index.names
    )


@doc(
    index_or_data="""
    index_or_data : DataFrame, Series or Index
        Index, Series or DataFrame of which to describe index levels\
    """
)
def describelevel(
    index_or_data: Union[DataFrame, Series, Index], n: int = 80, as_str: bool = False
) -> Optional[str]:
    """Describe index levels.

    Parameters
    ----------\
    {index_or_data}
    n : int, default 80
        The maximum line length
    as_str : bool, default False
        Whether to return as string or print, instead

    Returns
    -------
    description : str, optional
        if as_str is True

    See also
    --------
    pandas.DataFrame.describe
    """
    if isinstance(index_or_data, DataFrame):
        index_desc = _describelevel(index_or_data.index, n=n)
        columns_desc = _describelevel(index_or_data.columns, n=n)
        description = f"Index:\n{index_desc}\n\nColumns:\n{columns_desc}"
    else:
        if isinstance(index_or_data, Series):
            index_or_data = index_or_data.index
        description = f"Index:\n{_describelevel(index_or_data, n=n)}"

    if as_str:
        return description

    print(description)


summarylevel = deprecated(
    describelevel, reason="Use describelevel instead", version="v0.2.5"
)


def index_names(s, raise_on_index=False):
    if isinstance(s, (Series, DataFrame)):
        s = s.index

    if isinstance(s, MultiIndex):
        names = s.names
    elif isinstance(s, Index):
        if raise_on_index:
            exc = (
                raise_on_index
                if raise_on_index is not True
                else RuntimeError("s must have a MultiIndex")
            )
            raise exc
        names = FrozenList([s.name])
    else:
        names = FrozenList(s)

    return names


def _ensure_multiindex(s: Index) -> MultiIndex:
    if isinstance(s, MultiIndex):
        return s
    return MultiIndex.from_arrays([s], names=[s.name])


def ensure_multiindex(s: T) -> T:
    if isinstance(s, Index):
        return _ensure_multiindex(s)

    return s.set_axis(_ensure_multiindex(s.index))


@doc(
    frame_or_series="""
    frame_or_series : DataFrame or Series
        Data to be filtered\
    """
)
def semijoin(
    frame_or_series: S,
    other: Index,
    *,
    how: Literal["left", "right", "inner", "outer"] = "left",
    level: Union[str, int, None] = None,
    sort: bool = False,
    axis: Axis = 0,
    fill_value: Any = no_default,
    fail_on_reorder: bool = False,
) -> S:
    """Semijoin *frame_or_series* by index *other*.

    Parameters
    ----------\
    {frame_or_series}
    other : Index
        Other index to join with
    how : {{'left', 'right', 'inner', 'outer'}}
        Join method to use
    level : None or str or int or
        Single level on which to join, if not given join on all
    sort : bool, optional
        Whether to sort the index
    axis : {{0, 1, "index", "columns"}}
        Axis on which to join
    fill_value
        Value for filling gaps introduced by right or outer joins
    fail_on_reorder : bool, default False
        Raise ValueError if index order cannot be guaranteed

    Returns
    -------
    DataFrame or Series

    Raises
    ------
    ValueError
        If *fail_on_reorder* is True and the new index order does not correspond
        to the order of other
    ValueError
        If *axis* is not 0, "index" or 1, "columns"
    TypeError
        if *frame_or_series* does not derive from DataFrame or Series

    See also
    --------
    pandas.Index.join
    """

    index = get_axis(frame_or_series, axis)

    if level is None:
        index = ensure_multiindex(index)
        other = ensure_multiindex(other)

    new_index, left_idx, right_idx = index.join(
        other, how=how, level=level, return_indexers=True, sort=sort
    )

    if fail_on_reorder and not (
        right_idx is None or (right_idx == np.arange(len(right_idx), dtype=int)).all()
    ):
        raise ValueError(
            "Given index was re-sorted. To avoid this, sort the index before."
        )

    if left_idx is None:
        return frame_or_series.set_axis(new_index, axis=axis)

    any_missing = not (left_idx != -1).all()
    if isinstance(frame_or_series, DataFrame):
        if axis in (0, "index"):
            data = frame_or_series.iloc[left_idx]
            index = data.index
        elif axis in (1, "columns"):
            data = frame_or_series.iloc[:, left_idx]
            index = data.columns
        if any_missing:
            data = data.where(
                pd.Series(left_idx != -1, index), other=fill_value, axis=axis
            )
    elif isinstance(frame_or_series, Series):
        data = frame_or_series.iloc[left_idx]
        if any_missing:
            data = data.where(left_idx != -1, other=fill_value)
    else:
        raise TypeError(
            f"frame_or_series must derive from DataFrame or Series, but is {type(frame_or_series)}"
        )

    return data.set_axis(new_index, axis=axis)


@doc(
    index_or_data="""
    index_or_data : Index or DataFrame or Series
        Data to be filtered\
    """
)
def antijoin(
    index_or_data: S,
    other: Index,
    *,
    level: Union[str, int, None] = None,
    axis: Axis = 0,
) -> S:
    """Antijoin *index_or_data* with index *other*.

    ie remove all occurrences of other from data

    Parameters
    ----------\
    {index_or_data}
    other : Index
        Other index to join with
    level : None or str or int or
        Single level on which to join, if not given join on all
    axis : {{0, 1, "index", "columns"}}
        Axis on which to join

    Returns
    -------
    Index or DataFrame or Series

    Raises
    ------
    ValueError
        If axis is not 0, "index" or 1, "columns"
    TypeError
        if index_or_data does not derive from DataFrame or Series

    See also
    --------
    pandas.Index.join
    """

    index = get_axis(index_or_data, axis)

    _, left_idx, right_idx = index.join(
        other, how="left", level=level, return_indexers=True
    )

    if isinstance(index_or_data, DataFrame):
        if axis in (0, "index"):
            getter = lambda s: index_or_data.iloc[s]
        elif axis in (1, "columns"):
            getter = lambda s: index_or_data.iloc[:, s]
    elif isinstance(index_or_data, Series):
        getter = lambda s: index_or_data.iloc[s]
    elif isinstance(index_or_data, Index):
        getter = lambda s: index_or_data[s]
    else:
        raise TypeError(
            f"frame_or_series must derive from DataFrame or Series or Index, but is {type(index_or_data)}"
        )

    if right_idx is None:
        return getter(slice(0, 0))

    if left_idx is None:
        return getter(right_idx == -1)

    return getter(left_idx[right_idx == -1])


def _extractlevel(
    index: Index,
    template: Optional[str] = None,
    keep: bool = False,
    regex: bool = False,
    **templates: str,
) -> Tuple[Index, List[str]]:
    index = ensure_multiindex(index)
    all_identifiers = set()

    if template is not None:
        if len(index.names) > 1:
            raise ValueError("``template`` may only be non-null for single index level")
        templates[index.names[0]] = template

    for dim, template in templates.items():
        if dim not in index.names:
            raise ValueError(f"{dim} not a dimension of index: {index.names}")

        levelnum = index.names.index(dim)
        labels = index.levels[levelnum]
        codes = index.codes[levelnum]

        if regex:
            regex_pattern = re.compile(f"^{template}$")
            identifiers = list(regex_pattern.groupindex)
        else:
            identifiers = re.findall(r"\{([a-zA-Z_]+)\}", template)
            regex_pattern = reduce(
                lambda s, ident: s.replace(rf"\{{{ident}\}}", rf"(?P<{ident}>.*?)"),
                identifiers,
                re.escape(template),
            )
            regex_pattern = re.compile(f"^{regex_pattern}$")

        components = labels.str.extract(regex_pattern, expand=True)

        all_identifiers.update(identifiers)
        index = assignlevel(
            index, **{ident: components[ident].values[codes] for ident in identifiers}
        )

    if not keep:
        index = index.droplevel(list(set(templates) - all_identifiers))

    return index, list(all_identifiers)


@doc(
    index_or_data="""
    index_or_data : DataFrame, Series or Index
        Data to modify\
    """
)
def extractlevel(
    index_or_data: T,
    template: Optional[str] = None,
    *,
    keep: bool = False,
    dropna: bool = True,
    regex: bool = False,
    drop: Optional[bool] = None,
    axis: Axis = 0,
    **templates: str,
) -> T:
    """Extract new index levels with *templates* matched against any index
    level.

    The ``**templates`` argument defines pairs of level names and templates.
    Given level names are matched against the template, f.ex. ``"Emi|{{gas}}|{{sector}}"``.
    Patterns (``{{gas}}`` or ``{{sector}}``) appearing in the template are extracted
    from the successful matches and added as new levels.

    .. versionchanged:: 0.5.0
        *drop* replaced by *keep* and default changed to not keep.
        *regex* added.

    Parameters
    ----------\
    {index_or_data}
    template : str, optional
        Extraction template for a single level
    keep : bool, default False
        Whether to keep the split dimension
    dropna : bool, default True
        Whether to drop the non-matching levels
    regex : bool, default False
        Whether templates are given as regular expressions
        (regexes must use named captures)
    axis : {{0, 1, "index", "columns"}}, default 0
        Axis of DataFrame to extract from
    drop : bool, optional
        Deprecated argument, use keep instead
    **templates : str
        Templates for splitting one or multiple levels

    Returns
    -------
    Index, Series or DataFrame

    Raises
    ------
    ValueError
        If ``dim`` is not a dimension of ``index_or_series``
    ValueError
        If ``template`` is given, while index has more than one level

    Examples
    --------
    >>> s = Series(
    ...     range(4),
    ...     MultiIndex.from_arrays(
    ...         [
    ...             ["SE|Elec|Bio", "SE|Elec|Coal", "PE|Coal", "SE|Elec"],
    ...             ["GWh", "GWh", "EJ", "GWh"],
    ...         ],
    ...         names=["variable", "unit"],
    ...     ),
    ... )
    >>> s
    variable      unit
    SE|Elec|Bio   GWh     0
    SE|Elec|Coal  GWh     1
    PE|Coal       EJ      2
    SE|Elec       GWh     3
    dtype: int64
    >>> extractlevel(s, variable="SE|{{type}}|{{fuel}}", keep=True)
    variable      unit  type  fuel
    SE|Elec|Bio   GWh   Elec  Bio     0
    SE|Elec|Coal  GWh   Elec  Coal    1
    dtype: int64

    >>> extractlevel(s, variable="SE|{{type}}|{{fuel}}")
    unit  type  fuel
    GWh   Elec  Bio     0
    GWh   Elec  Coal    1
    dtype: int64

    >>> extractlevel(s, variable="SE|{{type}}|{{fuel}}", keep=True, dropna=False)
    variable      unit  type  fuel
    SE|Elec|Bio   GWh   Elec  Bio     0
    SE|Elec|Coal  GWh   Elec  Coal    1
    PE|Coal       EJ    NaN   NaN     2
    SE|Elec       GWh   NaN   NaN     3
    dtype: int64

    >>> extractlevel(s, variable=r"SE\\|(?P<type>.*?)(?:\\|(?P<fuel>.*?))?", regex=True)
    unit  type  fuel
    GWh   Elec  Bio     0
    GWh   Elec  Coal    1
    GWh   Elec  NaN     3
    dtype: int64

    >>> s = Series(range(3), ["SE|Elec|Bio", "SE|Elec|Coal", "PE|Coal"])
    >>> extractlevel(s, "SE|{{type}}|{{fuel}}")
    type  fuel
    Elec  Bio     0
          Coal    1
    dtype: int64

    See also
    --------
    formatlevel
    """
    if drop is not None:
        warnings.warn(
            "Argument `drop` is deprecated (use `keep` instead)", DeprecationWarning
        )
        keep = not drop

    if isinstance(index_or_data, Index):
        index_or_data, identifiers = _extractlevel(
            index_or_data, template, keep=keep, regex=regex, **templates
        )
    else:
        index, identifiers = _extractlevel(
            get_axis(index_or_data, axis), template, keep=keep, regex=regex, **templates
        )
        index_or_data = index_or_data.set_axis(index, axis=axis)

    if dropna:
        index_or_data = dropnalevel(
            index_or_data, subset=identifiers, how="all", axis=axis
        )

    return index_or_data


def _formatlevel(index: Index, drop: bool = False, **templates: str) -> Index:
    levels = {}
    used_levels = set()
    for dim, template in templates.items():
        # Build string
        string = ""
        prev_end = 0
        for m in re.finditer(r"\{([a-zA-Z_]+)\}", template):
            level = m.group(1)
            start, end = m.span()
            string += template[prev_end:start] + projectlevel(index, level).astype(str)
            prev_end = end
            used_levels.add(level)
        string += template[prev_end:]

        levels[dim] = string

    used_levels.difference_update(templates)

    if drop and used_levels:
        if len(used_levels) == len(index.levels):
            # none remain
            def expand_to_array(val):
                return (
                    val if not np.isscalar(val) else np.full(len(index), fill_value=val)
                )

            if len(levels) == 1:
                name, val = levels.popitem()
                return Index(expand_to_array(val), name=name)

            return MultiIndex.from_arrays(
                map(expand_to_array, levels.values()), names=levels.keys()
            )

        index = index.droplevel(list(used_levels))

    return assignlevel(index, **levels)


@doc(
    index_or_data="""
    index_or_data : DataFrame, Series or Index
        Data to modify\
    """
)
def formatlevel(
    index_or_data: T,
    drop: bool = False,
    axis: Axis = 0,
    **templates: str,
) -> T:
    """Format index levels based on a *template* which can refer to other
    levels.

    Parameters
    ----------\
    {index_or_data}
    drop : bool, default False
        Whether to drop the used index levels
    axis : {{0, 1, "index", "columns"}}, default 0
        Axis of DataFrame to modify
    **templates : str
        Format templates for one or multiple levels

    Returns
    -------
    Index, Series or DataFrame

    Raises
    ------
    ValueError
        If *templates* refer to non-existant levels
    """
    if isinstance(index_or_data, Index):
        return _formatlevel(index_or_data, drop, **templates)

    index = get_axis(index_or_data, axis)
    return index_or_data.set_axis(_formatlevel(index, drop, **templates), axis=axis)


def _fixindexna(index: Index):
    return index.set_codes(index.codes)


@doc(
    index_or_data="""
    index_or_data : Index, Series or DataFrame
        Data\
    """
)
def fixindexna(index_or_data: T, axis: Axis = 0) -> T:
    """Fix broken MultiIndex NA representation from .groupby(..., dropna=False)

    Refer to https://github.com/coroa/pandas-indexing/issues/25 for details

    Parameters
    ----------\
    {index_or_data}
    axis : Axis, optional
        Axis to fix, by default 0

    Returns
    -------
    index_or_data
    """
    if isinstance(index_or_data, Index):
        return _fixindexna(index_or_data)

    new_index = _fixindexna(get_axis(index_or_data, axis))
    return index_or_data.set_axis(new_index, axis=axis)


@doc(
    data="""
    data : Data
        Data in time-series representation with years on columns\
    """
)
def to_tidy(
    data: Data,
    meta: Optional[DataFrame] = None,
    value_name: Optional[str] = "value",
    columns: Optional[str] = "year",
) -> DataFrame:
    """Convert multi-indexed time-series dataframe to tidy dataframe.

    Parameters
    ----------\
    {data}
    meta : DataFrame, optional
        Meta data that is joined before tidying up
    value_name : str, optional
        Column name for the values; default "value"
        Use ``None`` to not change the name.
    columns : str, optional
        Name for the level on the columns axis; default "year"
        Use ``None`` to not change the name.

    Returns
    -------
    DataFrame
        Tidy dataframe without index
    """
    if isinstance(data, DataFrame):
        if columns is not None:
            data = data.rename_axis(columns=columns)
        data = data.stack()
    elif isinstance(data, Series):
        pass
    else:
        raise TypeError(
            f"data should be either a DataFrame or a Series, but found: {type(data)}"
        )
    if value_name is not None:
        data = data.rename(value_name)
    if meta is not None:
        data = data.to_frame().join(meta, on=meta.index.names)
    return data.reset_index()


def _aggregatelevel(
    data: T, agg_func: str = "sum", axis: Axis = 0, dropna: bool = True
):
    if axis in (0, "index"):
        return data.groupby(data.index.names, dropna=dropna).agg(agg_func)
    elif axis in (1, "columns"):
        return data.T.groupby(data.columns.names, dropna=dropna).agg(agg_func).T
    else:
        raise ValueError(
            f"axis can only be one of 0, 1, 'index' or 'columns', not: {axis}"
        )


@doc(
    data="""
    data : Data
        Series or DataFrame to aggregate\
    """
)
def aggregatelevel(
    data: T,
    agg_func: str = "sum",
    axis: Axis = 0,
    dropna: bool = True,
    mode: Literal["replace", "add", "return"] = "replace",
    **levels: Dict[str, Sequence[Any]],
) -> T:
    """Aggregate labels on one or multiple levels together.

    Parameters
    ----------\
    {data}
    agg_func : str, optional
        Function for aggregating values, default "sum"
        Other sensible options are "mean" or "first"
    axis : Axis, optional
        Axis on which to aggregate, default 0
    dropna : bool, optional
        Whether to drop or preserve NANs in the index, default True
    mode : {{"replace", "append", "return"}}
        Whether to replace or to append to the individual labels or return
        the aggregated data
    **levels
        Mapping for one or multiple levels, which labels to aggregate under a
        common name f.ex. ``region={{"sdn_ssd": ["sdn", "ssd"]}}`` aggregates
        the "sdn" and "ssd" regions to a new "sdn_ssd" region.

    Returns
    -------
    Data
        Aggregated data

    Notes
    -----
    If you already have a complete mapping from country to region, then prefer
    to use groupby directly instead of relying on this relatively slow method.

    See also
    --------
    pandas.DataFrame.groupby
    """

    if mode == "replace":
        for level, mapping in levels.items():
            data = data.rename(
                {
                    old_lbl: new_lbl
                    for new_lbl, old_lbls in mapping.items()
                    for old_lbl in old_lbls
                },
                axis=axis,
                level=level,
            )

        return _aggregatelevel(data, agg_func=agg_func, axis=axis, dropna=dropna)

    elif mode in ("append", "return"):
        new_data = [data]
        for level, mapping in levels.items():
            new_data.extend(
                assignlevel(
                    df.loc(axis=axis)[isin(**{level: old_lbls})],
                    **{level: new_lbl},
                    axis=axis,
                )
                for df, (new_lbl, old_lbls) in product(new_data, mapping.items())
            )

        new_data = _aggregatelevel(
            concat(new_data[1:], axis=axis), agg_func=agg_func, axis=axis, dropna=dropna
        )

        if mode == "return":
            return new_data

        # if any new label already exists, we combine otherwise we just concat
        def has_any_label(index: MultiIndex, level: str, labels: Sequence[Any]):
            levels = index.levels[index.names.index(level)]
            return not levels.intersection(labels).empty

        if any(
            has_any_label(get_axis(data, axis=axis), level, mapping.keys())
            for level, mapping in levels.items()
        ):
            return data.combine_first(new_data)
        else:
            return concat([data, new_data], axis=axis, sort=True)
    else:
        raise ValueError(
            f'mode must be "replace", "append" or "return", but is "{mode}"'
        )


@doc(
    data="""
    data : Data
        Series or DataFrame to extend with zeros\
    """
)
def add_zeros_like(
    data: T,
    reference: Union[MultiIndex, DataFrame, Series],
    *,
    derive: Optional[Dict[str, MultiIndex]] = None,
    **levels: Sequence[str],
) -> T:
    """Add explicit *levels* to *data* as 0 values.

    Remaining levels in *data* not found in *levels* or *derive* are taken from
    *reference* (or its index).

    Parameters
    ----------\
    {data}
    reference : Index
        expected level labels (like model, scenario combinations)
    derive : dict
        derive labels in a level from a multiindex with allowed combinations
    **levels : [str]
        which labels should be added to df

    Returns
    -------
    DataFrame
        unsorted data with additional zero data
    """

    if any(len(labels) == 0 for labels in levels.values()):
        return data

    if isinstance(reference, (Series, DataFrame)):
        reference = reference.index

    if derive is None:
        derive = {}

    target_levels = data.index.names
    index = reference.pix.unique(
        target_levels.difference(levels.keys()).difference(derive.keys())
    )

    zero_index = concat(
        reduce(
            lambda ind, d: ind.join(d, how="left"),
            derive.values(),
            index.pix.assign(**dict(zip(levels.keys(), labels))),
        ).reorder_levels(target_levels)
        for labels in product(*levels.values())
    )
    zero_index = antijoin(zero_index, data.index)

    return concat([data, pd.DataFrame(0, index=zero_index, columns=data.columns)])
